/*
 * Font handling functions (font.c)
 * Part of libpogo, a c-library replacent for GBA
 * Programmed by Jonas Minnberg (Sasq)
 *
 * DESCRIPTION
 * Functions for writing text with different fonts into
 * an 8bit VRAM mode.
 *
 **/

#include "core.h"
#include "device.h"
#include "font.h"

#include "screen.h"

/* Convert a 1bit bitmap to an 8bit bitmap */
static void planar2chunky(unsigned char *d, uint32 *s, int x, int y, int width, int height, int swidth, int dwidth)
{
	uint32 mask,m;
	uint32 startm = (0x80000000 >> (x%32));
	uint32 endm = (0x80000000 >> ((x+width)%32));
	uchar fg = 0x01;
	uchar bg = 0x00;
	int startx = x/32;
	int stopx = (x+width)/32;
	uint32 *sp = &s[startx + y*swidth];

	int smod = swidth - (stopx - startx) - 1;
	int dmod = dwidth - width;

	while(height--) {
		mask = startm;
		for(x = startx; x<=stopx; x++, sp++) {
			m = (x == stopx) ? endm : 0;
			for(; mask != m; mask /= 2)
				*d++ = (mask & *sp ? fg : bg);
			mask = 0x80000000;
		}

		d += dmod;
		sp += smod;
	}	
}

static Font *font_prepare(Font *font)
{
	if(!font->pixels)
		font->pixels = malloc(font->width * font->height);	
	planar2chunky(font->pixels, font->planar, 0, 0, font->width, font->height, (font->width+31)/32, font->width);
	return font;

}

static uchar font_putmono(Font *font, char c, int dest, int width)
{
	register int w;
	int solid = !(font->flags & FFLG_TRANSP);
	if(c < font->first || c > font->last)
		return 0;

	w = font->charwidth;
	if(dest != DONT_DRAW)
	{
		if(c == ' ')
		{
			if(solid)
				screen_clear_bg(screen_ptr(dest), w, font->height);
		}
		else
		{
			if(solid)
				screen_copy(screen_ptr(dest), &font->pixels[(c - font->first) * w], w, font->height, font->width);
			else
				screen_copy_fg(screen_ptr(dest), &font->pixels[(c - font->first) * w], w, font->height, font->width);
		}
	}
	return w;
}

void print(int x, int y, char *txt);

static uchar font_putprop(Font *font, char c, int dest, int width)
{
	register int offset,w;
	register int ff = font->first;
	int solid = !(font->flags & FFLG_TRANSP);

	if(c == ' ')
	{
		w = font->charwidth;
		if((dest != DONT_DRAW) && solid)
			screen_clear_bg(screen_ptr(dest), w, font->height);
	} 
	else
	{
		if(c > font->last)
			return 0;
		offset =  font->offsets[c - ff];
		w = font->offsets[c - ff + 1] - offset;
		if(dest != DONT_DRAW)
		{
			if(solid)
				screen_copy(screen_ptr(dest), &font->pixels[offset], w, font->height, font->width);
			else
				screen_copy_fg(screen_ptr(dest), &font->pixels[offset], w, font->height, font->width);
		}
	}
	return w;
}

uchar font_putchar(Font *font, char c, int dest, int width)
{
	int rc, fl;
	if(font->offsets)
	{
		if(font->flags & FFLG_BOLD)
		{
			font_putprop(font, c, dest, width);
			fl = font->flags;
			font->flags |= FFLG_TRANSP;
			rc = font_putprop(font, c, dest+1, width);
			font->flags = fl;
			return rc;
		}
		else
		return font_putprop(font, c, dest, width);
	}
	else
	{
		if(font->flags & FFLG_BOLD)
		{
			font_putmono(font, c, dest, width);
			fl = font->flags;
			font->flags |= FFLG_TRANSP;
			rc = font_putmono(font, c, dest+1, width);
			font->flags = fl;
			return rc;
		}
		else
		return font_putmono(font, c, dest, width);
	}
}

int font_text(Font *font, char *str, int dest)
{
	int outw = dest;

	while(*str)
		dest += XY(font_putchar(font, *str++, dest, 0), 0);
/*
	if(font->offsets)
		while(*str)
			dest += font_putprop(font, *str++, dest, width);
	else
		while(*str)
			dest += font_putmono(font, *str++, dest, width);
*/	
	return dest-outw;
}

int font_text_xy(Font *font, char *str, int x, int y)
{
	int dest = XY(x, y);
	return font_text(font, str, dest);
}


Font *font_load(char *name)
{
	int size, pix_size, l;
	Font font;
	Font *rfont;
	int fd = open(name, 0);
	if(fd > 0)
	{
		read(fd, &font, 8);
	
		pix_size = ((font.width+31)&0xFFFFFFE0) * font.height;

		l = strlen(name);
		if(!(l & 1))
			l++;
		size = pix_size + sizeof(Font) + l + 1;
		if(font.flags & FFLG_PROPORTIONAL)
			size += (font.last - font.first + 1)*2;

		rfont = malloc(size);
		memcpy(rfont, &font, 8);

		rfont->name = (char *)&rfont[1];
		strcpy(rfont->name, name);
		rfont->pixels = (uchar *)&rfont->name[l+1];
		rfont->planar = malloc(pix_size/8);
		read(fd, rfont->planar, pix_size/8);
		font_prepare(rfont);
		free(rfont->planar);
		rfont->planar = NULL;

		if(rfont->flags & FFLG_PROPORTIONAL)
		{
			rfont->offsets = (uint16 *)&rfont->pixels[pix_size];
			read(fd, rfont->offsets, (rfont->last - rfont->first + 1)*2);
		} else
			rfont->offsets = NULL;

		close(fd);

		return rfont;
	}
	return NULL;
}

Font *font_memload(uchar *data)
{
	int size, pix_size, l;
	Font font;
	Font *rfont;

	memcpy(&font, data, 8);
	
	pix_size = ((font.width+31)&0xFFFFFFE0) * font.height;

	l = 1;
	if(!(l & 1))
		l++;
	size = pix_size + sizeof(Font) + l + 1;
	if(font.flags & FFLG_PROPORTIONAL)
		size += (font.last - font.first + 1)*2;

	rfont = malloc(size);
	memcpy(rfont, &font, 8);

	rfont->name = (char *)&rfont[1];
	strcpy(rfont->name, "");
	rfont->pixels = (uchar *)&rfont->name[l+1];
	rfont->planar = malloc(pix_size/8);

	memcpy(rfont->planar, &data[8], pix_size/8);
	//read(fd, rfont->planar, pix_size/8);
	font_prepare(rfont);
	free(rfont->planar);
	rfont->planar = NULL;

	if(rfont->flags & FFLG_PROPORTIONAL)
	{
		rfont->offsets = (uint16 *)&rfont->pixels[pix_size];
		//read(fd, rfont->offsets, (rfont->last - rfont->first + 1)*2);
		memcpy(rfont->offsets, &data[8+pix_size/8], (rfont->last - rfont->first + 1)*2);
	} else
		rfont->offsets = NULL;

	return rfont;
}